"""Scripts to set the sample date udfs.
"""
import logging
from typing import Literal, List

import click
from genologics.entities import Sample, Process

from cg_lims import options
from cg_lims.get.samples import get_process_samples
from cg_lims.exceptions import MissingUDFsError
from datetime import datetime

LOG = logging.getLogger(__name__)


def set_prepared(sample: Sample) -> None:
    """Script to set todays date on sample udf Library Prep Finished.
    If the sample has a delivery finished date and a sequencing finished date when
    this script is run, they are deleted.
    This is because the sample is assumed to be re sequenced and delivered again."""

    if sample.udf.get("Delivered at"):
        sample.udf["Delivered at"] = None
        LOG.warning(
            f"The sample {sample.id} had a delivery finished date which is now deleted since the sample is being "
            f"re prepped. "
        )
    if sample.udf.get("Sequencing Finished"):
        LOG.warning(
            f"The sample {sample.id} had a sequencing finished date which is now deleted since the sample is "
            f"being re prepped. "
        )
        sample.udf["Sequencing Finished"] = None
    sample.udf["Library Prep Finished"] = datetime.today().date()
    sample.put()


def set_sequenced(sample: Sample) -> None:
    """Script to set todays date on sample udf Sequencing Finished.
    If the sample has a delivery finished date when this script is run, it will be deleted.
    This is because the sample is assumed to be delivered again."""

    if sample.udf.get("Sequencing Finished"):
        LOG.warning(f"The sample {sample.id} already had a sequencing date. Not updating.")
        return
    sample.udf["Sequencing Finished"] = datetime.today().date()
    sample.put()


def set_delivered(sample: Sample, process: Process) -> None:
    """Script to set delivery date on sample udf Delivered at.
    Overwriting any old delivery date."""

    date = process.udf.get("Date delivered")
    if not date:
        raise MissingUDFsError(message="Delivery date not set!")
    sample.udf["Delivered at"] = date
    sample.put()


@click.command()
@options.sample_udf(help="Sample date udf to set.")
@click.pass_context
def set_sample_date(
    context: click.Context,
    sample_udf: Literal[
        "Received at", "Library Prep Finished", "Sequencing Finished", "Delivered at"
    ],
):
    """Script to set date on sample udf."""

    LOG.info(f"Running {context.command_path} with params: {context.params}")
    process: Process = context.obj["process"]
    samples: List[Sample] = get_process_samples(process=process)
    for sample in samples:
        if sample_udf == "Library Prep Finished":
            set_prepared(sample=sample)
        elif sample_udf == "Sequencing Finished":
            set_sequenced(sample=sample)
        elif sample_udf == "Delivered at":
            set_delivered(sample=sample, process=process)



from typing import Optional

from pydantic import BaseModel


class PlacementMapHeader(BaseModel):
    process_type: str
    date: str


class PoolSection(BaseModel):
    nr_samples: int
    pool_name: str
    pool_id: str
    pools_information: str


class SampleTableSection(BaseModel):
    sample_id: str
    sample_warning_color: Optional[str]
    source_well: str
    source_container: str
    source_container_color: str
    pool_name: str
    extra_sample_values: str



from typing import Optional

from starlette import status


class LimsError(Exception):
    """Base Error"""

    def __init__(self, message):
        self.message = message


class AtlasResponseFailedError(LimsError):
    """Raise when failing retrieve atlas response."""

    pass


class ArgumentError(LimsError):
    """Raise when failing retrieve atlas response."""

    pass


class QueueArtifactsError(LimsError):
    """Raise when failing to route artifacts."""

    pass


class MissingFileError(LimsError):
    pass


class DuplicateSampleError(LimsError):
    """Raise when you excpect one sample, but find more.
    Eg: Two pools in a run contain the same sample."""

    pass


class MissingArtifactError(LimsError):
    """Raise when searching for artifacts that don't exist.
    Eg: Found no artifact for sample X in process Y."""

    pass


class MissingProcessError(LimsError):
    """Raise when searching for process that don't exist.."""

    pass


class MissingSampleError(LimsError):
    """Raise when searching for samples that don't exist."""

    pass


class MissingUDFsError(LimsError):
    """Raise when searching for udfs that don't exist.
    Eg: Found no udf X on artifact Y."""

    pass


class ZeroReadsError(LimsError):
    """Raise when read count is unexpectedly zero."""

    pass


class LowAmountError(LimsError):
    """Raise when amount is low."""

    pass


class FailingQCError(LimsError):
    """Raise when qc fails"""

    pass


class MissingCgFieldError(LimsError):
    """Raise when field missing"""

    pass


class FileError(LimsError):
    """Raise when error with file"""

    pass


class MissingValueError(LimsError):
    """Raise when value missing"""

    pass


class InvalidValueError(LimsError):
    """Raise when a value is invalid"""

    pass


class CSVColumnError(LimsError):
    """Raise when handling errors with csv columns"""

    pass


class InsertError(LimsError):
    def __init__(self, message: str, code: Optional[int] = status.HTTP_405_METHOD_NOT_ALLOWED):
        self.message = message
        self.code = code
        super().__init__(message)


class HighConcentrationError(LimsError):
    """Raise when a value is invalid"""

    pass



import logging
import sys

import click
from genologics.entities import Process
from genologics.lims import Lims

from cg_lims import options
from cg_lims.exceptions import LimsError, MissingUDFsError
from cg_lims.get.artifacts import get_artifacts

LOG = logging.getLogger(__name__)

def artifacts_to_sample(
    artifacts: list, sample_qc_udf: str = None
) -> None:

    """Function to copy artifact qc flag to sample UDF.

    For each artifact in the artifacts list, copying the qc_flagg
    to all samples related to the artifact. If a art is a pool, art.samples
    is a list with many samples. Otherwiese a list with only one sample.

    Arguments:
        artifacts: list of artifacts to copy from
        sample_qc_udf: sample qc udf to set based on artifact qc_flag"""

    failed_udfs = 0
    passed_udfs = 0
    for artifact in artifacts:
        if artifact.qc_flag != 'UNKNOWN' and artifact.qc_flag is not None:
            for sample in artifact.samples:
                if artifact.qc_flag == "PASSED":
                    sample.udf[sample_qc_udf] = "True"
                else :
                    sample.udf[sample_qc_udf] = "False"
                sample.put()
                passed_udfs += 1  
        else:
            LOG.error(
                f"Sample {artifact.id} is missing qc_flag"
            )
            failed_udfs += 1
    if failed_udfs:          
        raise MissingUDFsError(
            message=f"The qc_flag was is missing for {failed_udfs} artifacts. Udfs were set on {passed_udfs} samples."
        )
        
@click.command()
@options.sample_qc_udf(help="Name of sample QC udf to set.")
@options.input(
    help="Use this flag if you want copy udfs from input artifacts. Defaulte is output artifacts."
)
@click.pass_context
def qc_to_sample(ctx, input,sample_qc_udf):
    LOG.info(
        f"Running {ctx.command_path} with params: {ctx.params}"
    )
    process = ctx.obj["process"]
    try:
        artifacts = get_artifacts(process=process, input=input)
        artifacts_to_sample(
            artifacts=artifacts,
            sample_qc_udf=sample_qc_udf
        )
        message = "Udfs have been set on all samples."
        LOG.info(message)
        click.echo(message)
    except LimsError as e:
        LOG.error(e.message)
        sys.exit(e.message)



"""Scripts to set the sample date udfs.
"""
import logging
import sys
from typing import List, Optional

import click

import requests
from genologics.entities import Process

from cg_lims import options
from cg_lims.exceptions import LimsError, MissingUDFsError, AtlasResponseFailedError

LOG = logging.getLogger(__name__)


def get_path(document_title: str, process: Process, atlas_host: str) -> str:
    """Get method document path from Atlas"""

    response = requests.get(f"{atlas_host}/title/{document_title}/path")
    if response.status_code != 200:
        raise AtlasResponseFailedError(message=f"{response.status_code} : {response.text}")
    return response.json()


def get_document_paths(document_udfs: List[str], process: Process, host: str) -> List[str]:
    """Get method document paths from Atlas"""
    method_documents = []
    for udf in document_udfs:
        document_title: str = process.udf.get(udf)
        if not document_title or (document_title == "Manual"):
            LOG.warning(f"Process Udf: {udf} does not exist, was left empty or was set to Manual.")
            continue
        document_path = get_path(document_title, process, host)
        if document_path:
            method_documents.append(document_path)
    return method_documents


def set_methods_and_version(method_documents: List[str], process: Process, host: str) -> None:
    """Set Method Documents and Atlas Version"""

    response = requests.get(f"{host}/version")
    if response.status_code != 200:
        raise AtlasResponseFailedError(f"{response.status_code} : {response.text}")
    process.udf["Methods"] = " ,".join(method_documents)
    process.udf["Atlas Version"] = response.json()
    process.put()
    LOG.info("Method Documents and Atlas Version have been set.")
    click.echo("Method Documents and Atlas Version have been set.")


@click.command()
@options.process_udfs(help="List of Method Document UDFs.")
@click.pass_context
def method_document(ctx: click.Context, process_udfs: List[str]):
    """Script to set method documents and atlas version."""

    LOG.info(f"Running {ctx.command_path} with params: {ctx.params}")
    process: Process = ctx.obj["process"]
    atlas_host: str = ctx.obj["atlas_host"]
    try:
        method_documents = get_document_paths(
            document_udfs=process_udfs, host=atlas_host, process=process
        )
        if not method_documents:
            LOG.info(msg="Found no valid Method document UDFs in the step.")
            click.echo("Found no Method document UDFs in the step.")
        else:
            set_methods_and_version(
                process=process, host=atlas_host, method_documents=method_documents
            )
    except LimsError as e:
        sys.exit(e.message)



import json

import requests

from cg_lims.exceptions import LimsError


class StatusDBAPI(object):
    def __init__(self, url=None):
        self.url = url

    def apptag(self, tag_name, key=None, entry_point="/applications"):
        try:
            res = requests.get(self.url + entry_point + "/" + tag_name)
            if key:
                return json.loads(res.text)[key]
            else:
                return json.loads(res.text)
        except ConnectionError:
            raise LimsError(message="No connection to clinical-api!")



import click


def config(help: str = "config file for lims connection.") -> click.option:
    return click.option("-c", "--config", required=True, help=help)


def process(help: str = "Lims id for current Process.") -> click.option:
    return click.option("-p", "--process", required=True, help=help)


def log(
    help: str = "Path to log file.",
) -> click.option:
    return click.option("-l", "--log", required=True, help=help)


def workflow_id(help: str = "Workflow id.") -> click.option:
    return click.option("-w", "--workflow-id", required=True, help=help)


def stage_id(help: str = "Stage id.") -> click.option:
    return click.option("-s", "--stage-id", required=True, help=help)


def udf(
    help: str = "UDF name",
) -> click.option:
    return click.option("-u", "--udf", required=False, help=help)


def buffer_udf(
    help: str = "UDF name",
) -> click.option:
    return click.option("-b", "--buffer-udf", required=True, help=help)


def volume_udf(
    help: str = "UDF name",
) -> click.option:
    return click.option("-v", "--volume-udf", required=True, help=help)


def well_field(
    help: str = "Well field in file",
) -> click.option:
    return click.option("-wf", "--well-field", required=True, help=help)


def value_field(
    help: str = "Value field in file",
) -> click.option:
    return click.option("-vf", "--value-field", required=True, help=help)


def sample_qc_udf(
    help: str = "UDF name",
) -> click.option:
    return click.option("-qc", "--sample-qc-udf", default="", required=False, help=help)


def artifact_udf(
    help: str = "UDF name",
) -> click.option:
    return click.option("-audf", "--artifact-udf", required=True, help=help)


def sample_udf(
    help: str = "UDF name",
) -> click.option:
    return click.option("-sudf", "--sample-udf", required=True, help=help)


def process_udf(
    help: str = "UDF name",
) -> click.option:
    return click.option("-pudf", "--process-udf", required=True, help=help)


def file_placeholder(
    help: str = "File.",
) -> click.option:
    return click.option("-f", "--file", required=False, help=help)


def samples_file(help: str = "Txt file with sample ids") -> click.option:
    return click.option(
        "--samples-file",
        required=False,
        help=help,
    )


def local_file(help="local file path for debug purposes.") -> click.option:
    return click.option("-lf", "--local_file", required=False, help=help)


def input(
    help: str = "Use this flag if you run the script from a QC step.",
) -> click.option:
    return click.option(
        "-i",
        "--input",
        default=False,
        is_flag=True,
        help=help,
    )


def qc_flag(
    help: str = "Use this if you want to copy cq_flag.",
) -> click.option:
    return click.option(
        "-qc",
        "--qc-flag",
        default=False,
        is_flag=True,
        help=help,
    )


def sample_artifact(
    help: str = "Use this flag if you want to get the sample artifact (the very first artifact that is created when a "
    "sample is put into the first step in workflow.).",
) -> click.option:
    return click.option(
        "-sa",
        "--sample-artifact",
        default=False,
        is_flag=True,
        help=help,
    )


def original_well(
    help: str = "Use this flag if you want original well instead of source well.",
) -> click.option:
    return click.option(
        "-o",
        "--original-well",
        default=False,
        is_flag=True,
        help=help,
    )


def measurement(
    help: str = "Use this flag if you run the script from a QC step and want "
    "to get the measurement artifacts.",
) -> click.option:
    return click.option(
        "-m",
        "--measurement",
        default=False,
        is_flag=True,
        help=help,
    )


def process_types(help: str = "Process type name.") -> click.option:
    return click.option(
        "-n",
        "--process-types",
        required=False,
        multiple=True,
        help=help,
    )


def artifact_udfs(help: str = "Artifact udfs.") -> click.option:
    return click.option(
        "-au",
        "--artifact-udfs",
        required=False,
        multiple=True,
        help=help,
    )


def source_artifact_udfs(help: str = "Artifact udfs.") -> click.option:
    return click.option(
        "-sau",
        "--source-artifact-udfs",
        required=False,
        multiple=True,
        help=help,
    )


def sample_udfs(help: str = "Sample udfs.") -> click.option:
    return click.option(
        "-su",
        "--sample-udfs",
        required=False,
        multiple=True,
        help=help,
    )


def pool_udfs(help: str = "Pool udfs.") -> click.option:
    return click.option(
        "-pu",
        "--pool-udfs",
        required=False,
        multiple=True,
        help=help,
    )


def process_udfs(help: str = "Process udfs.") -> click.option:
    return click.option(
        "-pru",
        "--process-udfs",
        required=False,
        multiple=True,
        help=help,
    )


def pooling_step(
    help: str = "True if run from a pooling step",
) -> click.option:
    return click.option(
        "-p",
        "--pooling-step",
        default=False,
        is_flag=True,
        help=help,
    )


def size_udf(help: str = "Udf for fetching size.") -> click.option:
    return click.option(
        "--size-udf",
        required=True,
        help=help,
    )


def concantration_udf(help: str = "Udf for Concantration (??).") -> click.option:
    return click.option(
        "--conc-udf",
        required=True,
        help=help,
    )


def concantration_nm_udf(help: str = "Udf for Concantration (nM).") -> click.option:
    return click.option(
        "--conc-nm-udf",
        required=True,
        help=help,
    )


def file_extension(help: str = "Define file extension") -> click.option:
    return click.option("-e", "--extension", required=False, default="", help=help)


def file_suffix(help: str = "Define file name suffix") -> click.option:
    return click.option("--suffix", required=False, default="", help=help)


def amount_udf_option(
    help: str = "String of UDF used to get amount value",
) -> click.option:
    return click.option("--amount-udf", required=True, help=help)


def volume_udf_option(
    help: str = "String of UDF used to get volume value",
) -> click.option:
    return click.option("--volume-udf", required=True, help=help)


def concentration_udf_option(
    help: str = "String of UDF used to get concentration value",
) -> click.option:
    return click.option("--concentration-udf", required=True, help=help)


def prep(help: str = "Prep type") -> click.option:
    return click.option(
        "--prep-type",
        required=True,
        help=help,
        type=click.Choice(["wgs", "twist", "micro", "cov", "rna"]),
    )


def sequencing_method(help: str = "Sequencing Method") -> click.option:
    return click.option(
        "--sequencing-method",
        required=True,
        help=help,
        type=click.Choice(["novaseq"]),
    )


def subtract_volume_option(
    help: str = "Subtracts volume taken from samples in QC checks",
) -> click.option:
    return click.option("--subtract-volume", type=click.Choice(["0", "3"]), default="3", help=help)


def automation_string(
    help: str = "Search string for automation",
) -> click.option:
    return click.option("--automation-string", required=True, help=help)


def lower_threshold(help: str = "Set lower threshold value") -> click.option:
    return click.option("-lt", "--lower-threshold", required=False, help=help, type=float)


def upper_threshold(help: str = "Set upper threshold value") -> click.option:
    return click.option("-ut", "--upper-threshold", required=False, help=help, type=float)


def ignore_fail(
    help: str = "Use this flag if you don't want to raise exception errors",
) -> click.option:
    return click.option(
        "--ignore-fail",
        default=False,
        is_flag=True,
        help=help,
    )
  
  
def sample_volume_limit(help: str = "Set sample volume limit") -> click.option:
    return click.option("-svl", "--sample-volume-limit", required=False, help=help, type=float)


def keep_failed_flags(
    help: str = "Stops the overwriting of failed QC flags.",
) -> click.option:
    return click.option(
        "--keep-failed-flags",
        default=False,
        is_flag=True,
        help=help,
    )



import click

from .set_samples_reads_missing import set_reads_missing_on_new_samples
from .set_sample_date import set_sample_date
from .set_method import method_document


@click.group(invoke_without_command=True)
@click.pass_context
def set(context: click.Context):
    """Main entry point of set commands"""
    pass


set.add_command(set_reads_missing_on_new_samples)
set.add_command(set_sample_date)
set.add_command(method_document)



from typing import Optional

from genologics.lims import Lims
from pydantic.main import BaseModel
from pydantic import Field

from cg_lims.models.arnold.base_step import BaseStep
from cg_lims.objects import BaseAnalyte


class ProcessUDFs(BaseModel):
    """Xp Denature & ExAmp (Nova Seq)"""

    dpx1_volume: Optional[str] = Field(None, alias="DPX1 Volume (ul)")
    dpx2_volume: Optional[str] = Field(None, alias="DPX2 Volume (ul)")
    dpx3_volume: Optional[str] = Field(None, alias="DPX3 Volume (ul)")
    tris_lot_nr: Optional[str] = Field(None, alias="Tris-HCl lot nr")
    naoh_lot_nr: Optional[str] = Field(None, alias="NaOH lot nr")
    phix_lot_nr: Optional[str] = Field(None, alias="PhiX lot nr")
    examp_lot_nr: Optional[str] = Field(None, alias="ExAmp lot nr")


class ArtifactUDFs(BaseModel):
    bp_aliquot_volume: Optional[float] = Field(None, alias="BP Aliquot Volume (ul)")
    naoh_volume: Optional[float] = Field(None, alias="NaOH Volume (ul)")
    tris_volume: Optional[float] = Field(None, alias="Tris-HCl Volume (ul)")
    phix_volume: Optional[float] = Field(None, alias="PhiX Volume (ul)")
    mastermix_per_lane: Optional[float] = Field(None, alias="Mastermix per Lane (ul)")


class ArnoldStep(BaseStep):
    process_udfs: ProcessUDFs
    artifact_udfs: ArtifactUDFs

    class Config:
        allow_population_by_field_name = True


def get_xp_denature_and_examp(lims: Lims, sample_id: str, prep_id: str) -> Optional[ArnoldStep]:
    analyte = BaseAnalyte(
        lims=lims,
        sample_id=sample_id,
        process_type="Xp Denature & ExAmp (Nova Seq)",
        optional_step=True,
    )
    if not analyte.process:
        return None
    return ArnoldStep(
        **analyte.base_fields(),
        process_udfs=ProcessUDFs(**analyte.process_udfs()),
        artifact_udfs=ArtifactUDFs(**analyte.artifact_udfs()),
        sample_id=sample_id,
        prep_id=prep_id,
        step_type="xp_denature_and_examp",
        workflow="NovaSeq"
    )



from tests.conftest import server
from pathlib import Path

from genologics.entities import Artifact

import pytest

from cg_lims.EPPs.files.hamilton.make_kapa_csv import get_file_data_and_write
from cg_lims.exceptions import MissingArtifactError, MissingUDFsError


def test_get_file_data_and_write_KAPA_library_preparation(kapa_library_preparation_file, lims):
    # GIVEN: A file name. and a lims with two samples that has been run
    # through a process of type <amount_step>. There  they got udf: Amount needed (ng) set.
    # The two samples are now in another step where they have location and
    # reagent labels set.
    # (See entety_json_data for details)
    server("test_make_kapa_csv")
    file_name = "some_file_name"
    amount_step = "Aliquot samples for enzymatic fragmentation TWIST v2"
    file = Path(file_name)
    curent_step_artifacts = [Artifact(lims, id="2-1155129"), Artifact(lims, id="2-1155130")]

    # WHEN running get_file_data_and_write
    get_file_data_and_write(lims, amount_step, curent_step_artifacts, file)

    # THEN the file has been created and the file content is as expected.
    # See more details in PR:
    # https://github.com/Clinical-Genomics/cg_lims/pull/13#issuecomment-716526239
    assert file.read_text() == kapa_library_preparation_file
    file.unlink()


def test_get_file_data_and_write_Enzymatic_fragmentation(enzymatic_fragmentation_file, lims):
    # GIVEN: A file name. and a lims with two samples that has been run
    # through a process of type <amount_step>. There  they got udf: Amount needed (ng) set.
    # The two samples are now in another step where they have location but no
    # reagent labels set.
    # (See entety_json_data for details)
    server("test_make_kapa_csv")
    file_name = "some_file_id"
    amount_step = "Aliquot samples for enzymatic fragmentation TWIST v2"
    file = Path(file_name)
    curent_step_artifacts = [Artifact(lims, id="2-1155124"), Artifact(lims, id="2-1155125")]

    # WHEN running get_file_data_and_write
    get_file_data_and_write(lims, amount_step, curent_step_artifacts, file)

    # THEN the file has been created and the file content is as expected.
    # See more details in PR:
    # https://github.com/Clinical-Genomics/cg_lims/pull/13#issuecomment-716526239
    assert file.read_text() == enzymatic_fragmentation_file
    file.unlink()


def test_get_file_data_and_write_missing_udf(kapa_library_preparation_file_missing_udf, lims):
    # GIVEN: A file name. and a lims with two samples that has been run
    # through a process of type <amount_step>. There one sample did not get the udf: Amount needed (ng) set.
    # The two samples are now in another step where they have location and
    # reagent labels set.
    # (See entety_json_data for details)
    server("make_kapa_csv_missing_udfs")
    file_name = "some_file_name"
    amount_step = "Aliquot samples for enzymatic fragmentation TWIST v2"
    file = Path(file_name)
    curent_step_artifacts = [Artifact(lims, id="2-1155129"), Artifact(lims, id="2-1155130")]

    # WHEN running get_file_data_and_write
    # THEN MissingUDFsError is being raised, but the file is still created for the sample that
    # did have the amount udf set.
    with pytest.raises(MissingUDFsError):
        get_file_data_and_write(lims, amount_step, curent_step_artifacts, file)
    assert file.read_text() == kapa_library_preparation_file_missing_udf
    file.unlink()


def test_get_file_data_and_write_missing_artifact(kapa_library_preparation_file_missing_udf, lims):
    # GIVEN: A file name. and a lims with two samples that has NOT been run
    # through a process of type <amount_step>.
    # The two samples are now in another step where they have location and
    # reagent labels set.
    # (See entety_json_data for details)
    server("make_kapa_csv_missing_udfs")
    file_name = "some_file_name"
    amount_step = "Aliquot samples for enzymatic fragmentation TWIST v2"
    file = Path(file_name)
    curent_step_artifacts = [
        Artifact(lims, id="2-1155129_other_sample"),
        Artifact(lims, id="2-1155130_other_sample"),
    ]

    # WHEN running get_file_data_and_write
    # THEN MissingArtifactError is being raised, and no file is being created.
    with pytest.raises(MissingArtifactError):
        get_file_data_and_write(lims, amount_step, curent_step_artifacts, file)
    assert file.exists() == False



import logging
from typing import List

import click
from genologics.lims import Process, Lims, Artifact
import requests
from requests import Response

from cg_lims.exceptions import LimsError
from cg_lims.get.artifacts import get_output_artifacts

from cg_lims.models.arnold.flow_cell import FlowCell, Lane


LOG = logging.getLogger(__name__)


def build_flow_cell_document(process: Process, lanes: List[Artifact]) -> FlowCell:
    flowcell = FlowCell(**dict(process.udf.items()), date=process.date_run)

    for lane in lanes:
        if not lane.location:
            continue
        flowcell.lanes.append(Lane(name=lane.name, **dict(lane.udf.items())))

    return flowcell


@click.command()
@click.pass_context
def flow_cell(ctx):
    """Creating FlowCell documents from a run in the arnold flow_cell collection."""

    LOG.info(f"Running {ctx.command_path} with params: {ctx.params}")
    process: Process = ctx.obj["process"]
    lims: Lims = ctx.obj["lims"]
    arnold_host: str = ctx.obj["arnold_host"]
    lanes = get_output_artifacts(
        process=process, output_generation_type=["PerInput"], lims=lims, output_type="ResultFile"
    )
    flow_cell_document: FlowCell = build_flow_cell_document(process=process, lanes=lanes)
    response: Response = requests.post(
        url=f"{arnold_host}/flow_cell",
        headers={"Content-Type": "application/json"},
        data=flow_cell_document.json(exclude_none=True),
    )
    if not response.ok:
        LOG.info(response.text)
        raise LimsError(response.text)

    LOG.info("Arnold output: %s", response.text)
    click.echo("FlowCell document inserted to arnold database")



#!/usr/bin/env python
import click

# commands
from cg_lims.EPPs.files.file_to_udf import csv_well_to_udf
from cg_lims.EPPs.files.hamilton.base import hamilton
from cg_lims.EPPs.files.pooling_map.make_pooling_map import pool_map
from cg_lims.EPPs.files.placement_map.make_96well_placement_map import placement_map
from cg_lims.EPPs.files.csv_for_kapa_truble_shooting.csv_for_kapa_debug import trouble_shoot_kapa


@click.group(invoke_without_command=True)
@click.pass_context
def files(ctx):
    """Main entry point of file commands"""
    pass


files.add_command(csv_well_to_udf)
files.add_command(pool_map)
files.add_command(placement_map)
files.add_command(hamilton)
files.add_command(trouble_shoot_kapa)



from typing import Optional, List
from datetime import date, datetime

from pydantic import BaseModel, Field, validator


class Lane(BaseModel):
    name: Optional[str] = ""
    percent_aligned_r1: Optional[float] = Field(None, alias="% Aligned R1")
    percent_aligned_r2: Optional[float] = Field(None, alias="% Aligned R2")
    percent_bases_q30_r1: Optional[float] = Field(None, alias="% Bases >=Q30 R1")
    percent_bases_q30_r2: Optional[float] = Field(None, alias="% Bases >=Q30 R2")
    percent_error_rate_r1: Optional[float] = Field(None, alias="% Error Rate R1")
    percent_error_rate_r2: Optional[float] = Field(None, alias="% Error Rate R2")
    percent_phasing_r1: Optional[float] = Field(None, alias="% Phasing R1")
    percent_prephasing_r1: Optional[float] = Field(None, alias="% Prephasing R1")
    percent_prephasing_r2: Optional[float] = Field(None, alias="% Prephasing R2")
    percentpf_r1: Optional[float] = Field(None, alias="%PF R1")
    percentpf_r2: Optional[float] = Field(None, alias="%PF R2")
    cluster_density_r1: Optional[float] = Field(None, alias="Cluster Density (K/mm^2) R1")
    cluster_density_r2: Optional[float] = Field(None, alias="Cluster Density (K/mm^2) R2")
    intensity_cycle_1_r1: Optional[float] = Field(None, alias="Intensity Cycle 1 R1")
    intensity_cycle_1_r2: Optional[float] = Field(None, alias="Intensity Cycle 1 R2")
    reads_pf_millions_r1: Optional[int] = Field(None, alias="Reads PF (M) R1")
    reads_pf_millions_r2: Optional[int] = Field(None, alias="Reads PF (M) R2")
    yield_pf_giga_bases_r1: Optional[float] = Field(None, alias="Yield PF (Gb) R1")
    yield_pf_giga_bases_r2: Optional[float] = Field(None, alias="Yield PF (Gb) R2")
    percent_phasing_r2: Optional[float] = Field(None, alias="% Phasing R2")


class FlowCell(BaseModel):
    instrument: Optional[str] = Field(None, alias="instrument")
    date: Optional[date]
    done: Optional[bool] = Field(None, alias="Done")
    buffer_expiration_date: Optional[date] = Field(None, alias="Buffer Expiration Date")
    buffer_lot_number: Optional[str] = Field(None, alias="Buffer Lot Number")
    buffer_part_number: Optional[str] = Field(None, alias="Buffer Part Number")
    buffer_serial_barcode: Optional[str] = Field(None, alias="Buffer Serial Barcode")
    flow_cell_expiration_date: Optional[date] = Field(None, alias="Flow Cell Expiration Date")
    flow_cell_id: Optional[str] = Field(None, alias="Flow Cell ID")
    flow_cell_lot_number: Optional[str] = Field(None, alias="Flow Cell Lot Number")
    flow_cell_mode: Optional[str] = Field(None, alias="Flow Cell Mode")
    flow_cell_part_number: Optional[str] = Field(None, alias="Flow Cell Part Number")
    pe_cycle_kit: Optional[str] = Field(None, alias="PE Cycle Kit")
    pe_expiration_date: Optional[date] = Field(None, alias="PE Expiration Date")
    pe_lot_number: Optional[str] = Field(None, alias="PE Lot Number")
    pe_part_number: Optional[str] = Field(None, alias="PE Part Number")
    pe_serial_barcode: Optional[str] = Field(None, alias="PE Serial Barcode")
    run_id: Optional[str] = Field(None, alias="Run ID")
    sbs_cycle_kit: Optional[str] = Field(None, alias="SBS Cycle Kit")
    sbs_expiration_date: Optional[date] = Field(None, alias="SBS Expiration Date")
    sbs_lot_number: Optional[str] = Field(None, alias="SBS Lot Number")
    sbs_part_number: Optional[str] = Field(None, alias="SBS Part Number")
    sbs_serial_barcode: Optional[str] = Field(None, alias="SBS Serial Barcode")
    lanes: Optional[List[Lane]] = []

    @validator(
        "sbs_expiration_date",
        "pe_expiration_date",
        "flow_cell_expiration_date",
        "buffer_expiration_date",
        "date",
        always=True,
    )
    def vali_date(cls, v, values) -> Optional[str]:
        if isinstance(v, date):
            return datetime(v.year, v.month, v.day).__str__()
        return None

    class Config:
        allow_population_by_field_name = True



import mock
import pytest
from genologics.entities import Sample

from cg_lims.EPPs.udf.set.set_samples_reads_missing import (
    get_target_amount,
    set_reads_missing,
    set_reads_missing_on_sample,
)
from cg_lims.exceptions import LimsError, MissingUDFsError
from tests.conftest import server


@mock.patch("cg_lims.get.udfs.get_udf")
@mock.patch("cg_lims.status_db_api.StatusDBAPI")
def test_get_target_amount(mock_status_db, mock_get_udf, sample_1: Sample):
    # GIVEN a sample with a udf "Sequencing Analysis" containing an app
    # tag
    server("flat_tests")
    test_app_tag = "TESTAPPTAG"
    sample_1.udf["Sequencing Analysis"] = test_app_tag

    # WHEN getting the target amount of reads via the StatusDBAPI
    mock_get_udf.return_value = test_app_tag
    mock_status_db.apptag.return_value = 9_000_000
    result = get_target_amount(test_app_tag, mock_status_db)

    # THEN the target amount of reads should be retrieved via the StatusDBAPI
    assert result == 9_000_000
    assert mock_status_db.apptag.mock_calls == [
        mock.call(tag_name=test_app_tag, key="target_reads")
    ]


@mock.patch("cg_lims.status_db_api.StatusDBAPI")
@mock.patch("cg_lims.EPPs.udf.set.set_samples_reads_missing.get_target_amount")
@mock.patch("cg_lims.EPPs.udf.set.set_samples_reads_missing.get_udf")
def test_set_reads_missing_on_sample(
    mock_get_udf,
    mock_get_target_amount,
    mock_status_db,
    sample_1: Sample,
):
    # GIVEN A SAMPLE
    sample = sample_1

    # WHEN setting the missing reads on that sample_1 based on the target amount
    mock_get_target_amount.return_value = 9_000_000

    set_reads_missing_on_sample(sample, mock_status_db)

    # THEN the udf "Reads missing (M)" on that sample should be set correctly
    assert sample.udf["Reads missing (M)"] == 9


@mock.patch(
    "cg_lims.EPPs.udf.set.set_samples_reads_missing.set_reads_missing_on_sample"
)
@mock.patch("cg_lims.status_db_api.StatusDBAPI")
def test_set_reads_missing_one_sample(
    mock_status_db, mock_set_reads_missing_on_sample, sample_1: Sample
):
    # GIVEN a single sample
    samples = [sample_1]

    # WHEN setting the reads missing on that sample
    set_reads_missing(samples, status_db=mock_status_db)

    # THEN set_missing_reads should be called for that sample
    mock_set_reads_missing_on_sample.assert_called_with(sample_1, mock_status_db)


@mock.patch(
    "cg_lims.EPPs.udf.set.set_samples_reads_missing.set_reads_missing_on_sample"
)
@mock.patch("cg_lims.status_db_api.StatusDBAPI")
def test_set_reads_missing_multiple_samples(
    mock_status_db, mock_set_reads_missing_on_sample, sample_1: Sample, sample_2: Sample
):
    # GIVEN multiple of samples
    samples = [sample_1, sample_2]

    # WHEN setting the reads missing on those samples
    set_reads_missing(samples, status_db=mock_status_db)

    # THEN the missing reads should be set on both samples
    assert mock_set_reads_missing_on_sample.mock_calls == [
        mock.call(sample_1, mock_status_db),
        mock.call(sample_2, mock_status_db),
    ]


@mock.patch(
    "cg_lims.EPPs.udf.set.set_samples_reads_missing.set_reads_missing_on_sample"
)
@mock.patch("cg_lims.status_db_api.StatusDBAPI")
def test_set_reads_missing_one_sample_exception(
    mock_status_db,
    mock_set_reads_missing_on_sample,
    sample_1: Sample,
):
    # GIVEN a single sample
    samples = [sample_1]

    # WHEN setting the reads missing on that sample leads to an exception  being raised
    mock_set_reads_missing_on_sample.side_effect = MissingUDFsError("TEST MISSING UDF")
    with pytest.raises(LimsError) as error:
        set_reads_missing(samples, mock_status_db)

    # AND one failed sample should be counted, and it's id should be returned
    mock_set_reads_missing_on_sample.assert_called_with(sample_1, mock_status_db)
    assert (
        error.value.message
        == "Reads Missing (M) set on 0 sample(s), 1 sample(s) failed"
    )


@mock.patch(
    "cg_lims.EPPs.udf.set.set_samples_reads_missing.set_reads_missing_on_sample"
)
@mock.patch("cg_lims.status_db_api.StatusDBAPI")
def test_set_reads_missing_multiple_samples_exception_on_first_sample(
    mock_status_db,
    mock_set_reads_missing_on_sample,
    sample_1: Sample,
    sample_2: Sample,
):
    # GIVEN multiple samples
    samples = [sample_1, sample_2]

    # WHEN setting the reads missing on those samples and the second sample leads to an exception being raised
    mock_set_reads_missing_on_sample.side_effect = (
        None,
        MissingUDFsError("TEST MISSING UDF"),
    )
    with pytest.raises(LimsError) as error:
        set_reads_missing(samples, mock_status_db)

    # THEN setting the missing reads should be attempted for both samples
    assert mock_set_reads_missing_on_sample.mock_calls == [
        mock.call(sample_1, mock_status_db),
        mock.call(sample_2, mock_status_db),
    ]
    # AND one failed sample should be counted, and it's id should be returned
    assert (
        error.value.message
        == "Reads Missing (M) set on 1 sample(s), 1 sample(s) failed"
    )


@mock.patch(
    "cg_lims.EPPs.udf.set.set_samples_reads_missing.set_reads_missing_on_sample"
)
@mock.patch("cg_lims.status_db_api.StatusDBAPI")
def test_set_reads_missing_multiple_samples_exception_on_second_sample(
    mock_status_db,
    mock_set_reads_missing_on_sample,
    sample_1: Sample,
    sample_2: Sample,
):
    # GIVEN multiple samples
    samples = [sample_1, sample_2]
    assert sample_1.udf.get("Sequencing Analysis") is None

    # WHEN setting the reads missing on those samples and the first sample leads to an exception being raised

    mock_set_reads_missing_on_sample.side_effect = (
        MissingUDFsError("TEST MISSING UDF"),
        None,
    )
    with pytest.raises(LimsError) as error:
        set_reads_missing(samples, mock_status_db)

    # THEN setting the missing reads should be attempted for both samples
    assert mock_set_reads_missing_on_sample.mock_calls == [
        mock.call(sample_1, mock_status_db),
        mock.call(sample_2, mock_status_db),
    ]
    # AND one failed sample should be counted
    assert (
        error.value.message
        == "Reads Missing (M) set on 1 sample(s), 1 sample(s) failed"
    )


@mock.patch(
    "cg_lims.EPPs.udf.set.set_samples_reads_missing.set_reads_missing_on_sample"
)
@mock.patch("cg_lims.status_db_api.StatusDBAPI")
def test_set_reads_missing_multiple_samples_exception_on_both_samples(
    mock_status_db,
    mock_set_reads_missing_on_sample,
    sample_1: Sample,
    sample_2: Sample,
):
    # GIVEN multiple samples
    samples = [sample_1, sample_2]
    assert sample_1.udf.get("Sequencing Analysis") is None

    # WHEN setting the reads missing on those samples and both samples leads to an exception being raised
    mock_set_reads_missing_on_sample.side_effect = (
        MissingUDFsError("TEST MISSING UDF"),
        MissingUDFsError("TEST MISSING UDF"),
    )
    with pytest.raises(LimsError) as error:
        set_reads_missing(samples, mock_status_db)

    # THEN setting the missing reads should be attempted for both samples
    assert mock_set_reads_missing_on_sample.mock_calls == [
        mock.call(sample_1, mock_status_db),
        mock.call(sample_2, mock_status_db),
    ]
    # AND two failed sample should be counted, and their id's should be returned
    assert (
        error.value.message
        == "Reads Missing (M) set on 0 sample(s), 2 sample(s) failed"
    )



from genologics.entities import Process
from cg_lims.EPPs.udf.calculate.sum_missing_reads_in_pool import sum_reads_in_pool
from cg_lims.get.artifacts import get_artifacts
from tests.conftest import server



def test_sum_reads_pool(lims):
    # GIVEN: A process with pools, where the samples of each pool have the udf 'Reads missing (M)' set.

    server("missing_reads_pool")
    process = Process(lims, id="24-196210")
    artifacts = get_artifacts(process=process, input=False)

    # WHEN running sum_reads_in_pool
    passed_artifacts, failed_artifacts = sum_reads_in_pool(artifacts=artifacts)

    # THEN all pools passed and each pool udf: 'Missing reads Pool (M)' is the sum of ther sample reads.
    assert failed_artifacts == 0
    for artifact in artifacts:
        assert artifact.udf.get('Missing reads Pool (M)') == sum(sample.udf.get("Reads missing (M)") for sample in artifact.samples)



def test_sum_reads_pool_missing_udf(lims):
    # GIVEN: A process with pools, where the samples of each pool have the udf 'Reads missing (M)' set except one.

    server("missing_reads_pool")
    process = Process(lims, id="24-196210")
    artifacts = get_artifacts(process=process, input=False)
    a_sample_in_a_pool = artifacts[0].samples[0]
    a_sample_in_a_pool.udf["Reads missing (M)"] = None
    a_sample_in_a_pool.put()

    # WHEN running sum_reads_in_pool,
    passed_artifacts, failed_artifacts = sum_reads_in_pool(artifacts=artifacts)

    # THEN assert one artifact failed
    assert failed_artifacts == 1



import pytest
from genologics.entities import Artifact, Process

from cg_lims.exceptions import MissingUDFsError, MissingCgFieldError
from cg_lims.EPPs.udf.copy.qc_to_sample import artifacts_to_sample
from cg_lims.set.udfs import copy_artifact_to_artifact
from tests.conftest import server

from cg_lims.get.artifacts import get_artifacts


def test_udf_Passed_Initial_QC(lims):
    # GIVEN every other artifacts have True or False on udf: Passed Initial QC.
    udf = "Passed Initial QC"
    server("calculate_average_size_and_set_qc")
    process = Process(lims, id='24-297378')
    artifacts = get_artifacts(process=process)
    
    odd_or_even = 0 
    for artifact in artifacts:
        if int(odd_or_even % 2) == 0:
            artifact.qc_flag = "PASSED"

        else:
            artifact.qc_flag = "FAILED"
        
        artifact.put()
        odd_or_even += 1

    # When running function artifacts_to_sample.
    artifacts_to_sample(artifacts, udf)

    # THEN the corresponding sample will have the matching qc_flag.
    for artifact in artifacts:
        sample = artifact.samples[0]
        if artifact.qc_flag == 'PASSED':
            assert sample.udf[udf] == 'True'

        elif artifact.qc_flag == 'FAILED':
            assert sample.udf[udf] == 'False'


def test_udf_Passed_Library_QC(lims):
    # GIVEN every other artifacts have True or False on udf: Passed Library QC.
    udf = "Passed Library QC"
    server("calculate_average_size_and_set_qc")
    process = Process(lims, id='24-297378')
    artifacts = get_artifacts(process=process)

    odd_or_even = 0 
    for artifact in artifacts:
        if int(odd_or_even % 2) == 0:
            artifact.qc_flag = "PASSED"

        else:
            artifact.qc_flag = "FAILED"
        
        artifact.put()
        odd_or_even += 1

    # When running function artifacts_to_sample.
    artifacts_to_sample(artifacts, udf)

    # THEN the corresponding sample will have the matching qc_flag.
    for artifact in artifacts:
        sample = artifact.samples[0]
        if artifact.qc_flag == 'PASSED':
            assert sample.udf[udf] == 'True'

        elif artifact.qc_flag == 'FAILED':
            assert sample.udf[udf] == 'False'


def test_change_value(lims):
    # GIVEN sample already has a value on Udfs "Passed Library QC" and "Passed Initial QC".
    server("calculate_average_size_and_set_qc")
    process = Process(lims, id='24-297378')
    artifacts = get_artifacts(process=process)

    odd_or_even = 0 
    for artifact in artifacts:
        if int(odd_or_even % 2) == 0:
            artifact.qc_flag = "PASSED"

        else:
            artifact.qc_flag = "FAILED"
        
        artifact.put()
        odd_or_even += 1    

    udfs = ["Passed Library QC", "Passed Initial QC"]
    for udf in udfs:
        artifacts_to_sample(artifacts, udf)

    # WHEN running function artifacts_to_sample with artifacts with new updated qc_flag values.
    # This for-loop will change all qc_flags (e.g. if previous qc_flag value was True then the new is False).
    odd_or_even = 0 
    for artifact in artifacts:
        if int(odd_or_even % 2) == 0:
            artifact.qc_flag = "FAILED"

        else:
            artifact.qc_flag = "PASSED"
        
        artifact.put()
        odd_or_even += 1    

    udfs = ["Passed Library QC", "Passed Initial QC"]
    for udf in udfs:
        artifacts_to_sample(artifacts, udf)

    # THEN sample should change from old to new value.
    udfs = ["Passed Library QC", "Passed Initial QC"]
    for udf in udfs:
        for artifact in artifacts:
            sample = artifact.samples[0]
            if artifact.qc_flag == 'PASSED':
                assert sample.udf[udf] == 'True'

            elif artifact.qc_flag == 'FAILED':
                assert sample.udf[udf] == 'False' 


def test_pool(lims):
    # GIVEN pools and that every other pool have True or False on udf: Passed Library QC.
    udf = "Passed Library QC"
    server("missing_reads_pool")
    process = Process(lims, id="24-196210")
    artifacts = get_artifacts(process=process, input = True)

    odd_or_even = 0 
    for artifact in artifacts:
        if int(odd_or_even % 2) == 0:
            artifact.qc_flag = "PASSED"

        else:
            artifact.qc_flag = "FAILED"
        
        artifact.put()
        odd_or_even += 1

    # WHEN running function artifacts_to_sample.
    artifacts_to_sample(artifacts, udf)

    # THEN the corresponding sample will have the matching qc_flag
    for artifact in artifacts:
        sample = artifact.samples[0]
        if artifact.qc_flag == 'PASSED':
            assert sample.udf[udf] == 'True'
        elif artifact.qc_flag == 'FAILED':
            assert sample.udf[udf] == 'False'


def test_unknown_qc_flag(lims):
    # GIVEN samples with unknown qc_flag value.
    udf = "Passed Library QC"
    server("calculate_average_size_and_set_qc")
    process = Process(lims, id='24-297378')
    artifacts = get_artifacts(process=process, input = True)

    # WHEN running function artifacts_to_sample.
    # THEN MissingUDFsERROR should be raised.
    with pytest.raises(MissingUDFsError) as error_message:
        artifacts_to_sample(artifacts, udf)



import datetime
from typing import Optional
from pydantic import Field, validator
from pydantic.main import BaseModel
from datetime import date, datetime


def get_number_of_days(first_date: str, second_date: str) -> Optional[int]:
    """Get number of days between different iso formatted time stamps."""
    if first_date and second_date:
        first_datetime = datetime.strptime(first_date, "%Y-%m-%d %H:%M:%S")
        second_datetime = datetime.strptime(second_date, "%Y-%m-%d %H:%M:%S")
        time_span = second_datetime - first_datetime
        return time_span.days


class ArnoldSample(BaseModel):
    id: Optional[str] = Field(alias="_id")
    sample_id: str
    category: Optional[str]
    received_date: Optional[date] = Field(alias="Received at")
    delivery_date: Optional[date] = Field(alias="Delivered at")
    sequenced_date: Optional[date] = Field(alias="Sequencing Finished")
    prepared_date: Optional[date] = Field(alias="Library Prep Finished")
    collection_date: Optional[date] = Field(alias="Collection Date")
    sequenced_to_delivered: Optional[int]
    prepped_to_sequenced: Optional[int]
    received_to_prepped: Optional[int]
    received_to_delivered: Optional[int]
    initial_qc: Optional[str] = Field(alias="Passed Initial QC")
    library_qc: Optional[str] = Field(alias="Passed Library QC")
    prep_method: Optional[str] = Field(alias="Prep Method")
    application: Optional[str] = Field(alias="Sequencing Analysis")  # application_tag in vogue
    bait_set: Optional[str] = Field(alias="Bait Set")
    capture_kit: Optional[str] = Field(alias="Capture Library version")
    comment: Optional[str] = Field(alias="Comment")
    concentration: Optional[float] = Field(alias="Concentration (nM)")
    concentration_sample: Optional[float] = Field(alias="Sample Conc.")
    customer: Optional[str] = Field(alias="customer")
    data_analysis: Optional[str] = Field(alias="Data Analysis")
    data_delivery: Optional[str] = Field(alias="Data Delivery")
    elution_buffer: Optional[str] = Field(alias="Sample Buffer")
    extraction_method: Optional[str] = Field(alias="Extraction method")
    family_name: Optional[str] = Field(alias="familyID")
    formalin_fixation_time: Optional[float] = Field(alias="Formalin Fixation Time")
    index: Optional[str] = Field(alias="Index type")
    index_number: Optional[str] = Field(alias="Index number")
    lab_code: Optional[str] = Field(alias="Lab Code")
    organism: Optional[str] = Field(alias="Strain")  # strain in vogue
    organism_other: Optional[str] = Field(alias="Other species")
    original_lab: Optional[str] = Field(alias="Original Lab")
    original_lab_address: Optional[str] = Field(alias="Original Lab Address")
    pool: Optional[str] = Field(alias="pool name")
    post_formalin_fixation_time: Optional[float] = Field(alias="Post Formalin Fixation Time")
    pre_processing_method: Optional[str] = Field(alias="Pre Processing Method")
    priority: Optional[str] = Field(alias="priority")
    quantity: Optional[str] = Field(alias="Quantity")
    reference_genome: Optional[str] = Field(alias="Reference Genome Microbial")
    region: Optional[str] = Field(alias="Region")
    region_code: Optional[str] = Field(alias="Region Code")
    require_qcok: Optional[str] = Field(alias="Process only if QC OK")
    rml_plate_name: Optional[str] = Field(alias="RML plate name")
    selection_criteria: Optional[str] = Field(alias="Selection Criteria")
    sequencing_qc_pass: Optional[str] = Field(
        alias="Passed Sequencing QC"
    )  # sequencing_qc in vogue
    sex: Optional[str] = Field(alias="Gender")
    source: Optional[str] = Field(alias="Source")
    target_reads: Optional[float] = Field(alias="Reads missing (M)")
    tissue_block_size: Optional[str] = Field(alias="Tissue Block Size")
    tumour: Optional[str] = Field(alias="tumor")
    tumour_purity: Optional[int] = Field(alias="tumour purity")
    verified_organism: Optional[str] = Field(alias="Verified organism")
    volume: Optional[float] = Field(alias="Volume (uL)")
    well_position_rml: Optional[str] = Field(alias="RML well position")

    @validator("received_date", always=True)
    def get_received_date(cls, v, values) -> Optional[str]:
        if isinstance(v, date):
            return datetime(v.year, v.month, v.day).__str__()
        return None

    @validator("delivery_date", always=True)
    def get_delivery_date(cls, v, values) -> Optional[str]:
        if isinstance(v, date):
            return datetime(v.year, v.month, v.day).__str__()
        return None

    @validator("sequenced_date", always=True)
    def get_sequenced_date(cls, v, values) -> Optional[str]:
        if isinstance(v, date):
            return datetime(v.year, v.month, v.day).__str__()
        return None

    @validator("collection_date", always=True)
    def get_collection_date(cls, v, values) -> Optional[str]:
        if isinstance(v, date):
            return datetime(v.year, v.month, v.day).__str__()
        return None

    @validator("prepared_date", always=True)
    def get_prepared_date(cls, v, values) -> Optional[str]:
        if isinstance(v, date):
            return datetime(v.year, v.month, v.day).__str__()
        return None

    @validator("sequenced_to_delivered", always=True)
    def get_sequenced_to_delivered(cls, v, values) -> Optional[int]:
        return get_number_of_days(values.get("sequenced_date"), values.get("delivered_date"))

    @validator("prepped_to_sequenced", always=True)
    def get_prepped_to_sequenced(cls, v, values) -> Optional[int]:
        return get_number_of_days(values.get("prepared_date"), values.get("sequenced_date"))

    @validator("received_to_prepped", always=True)
    def get_received_to_prepped(cls, v, values) -> Optional[int]:
        return get_number_of_days(values.get("received_date"), values.get("prepared_date"))

    @validator("received_to_delivered", always=True)
    def get_received_to_delivered(cls, v, values) -> Optional[int]:
        return get_number_of_days(values.get("received_date"), values.get("delivered_date"))

    class Config:
        allow_population_by_field_name = True



from tests.conftest import server
from pathlib import Path

from genologics.entities import Artifact

import pytest

from cg_lims.EPPs.files.hamilton.buffer_exchange_twist_file import get_file_data_and_write
from cg_lims.exceptions import MissingUDFsError


def test_get_file_data_and_write(lims, hamilton_buffer_exchange):
    # GIVEN a file name and a list of artifacts with all required udf:s and assorted wells.
    server("buffer_exchange_twist_file")
    file_name = "file_name_1"
    file = Path(file_name)
    artifacts = [
        Artifact(lims, id="2-2336703"),
        Artifact(lims, id="2-2336704"),
        Artifact(lims, id="2-2336705"),
        Artifact(lims, id="2-2336706"),
        Artifact(lims, id="2-2336707"),
    ]

    # WHEN running get_file_data_and_write
    get_file_data_and_write(artifacts, file)

    # THEN the file has been created and the file content is as expected.
    assert file.read_text() == hamilton_buffer_exchange
    file.unlink()


def test_get_file_data_and_write_missing_udfs(lims, hamilton_buffer_exchange_no_udf):
    # GIVEN a file name and a list of artifacts where all but one has registered udf:s
    server("buffer_exchange_twist_file")
    file_name = "file_name_2"
    file = Path(file_name)
    artifacts = [
        Artifact(lims, id="2-2336703"),
        Artifact(lims, id="2-2336704"),
        Artifact(lims, id="2-2336705"),
        Artifact(lims, id="2-2336706"),
        Artifact(lims, id="2-2336707"),
        Artifact(lims, id="2-2336708"),
    ]

    # WHEN running get_file_data_and_write
    get_file_data_and_write(artifacts, file)

    # THEN the file has been created and missing udfs are replaced with the symbol "-".
    assert file.read_text() == hamilton_buffer_exchange_no_udf
    file.unlink()


def test_get_file_data_and_write_no_well(lims):
    # GIVEN a file name and one artifact which does not have an assigned well
    server("buffer_exchange_twist_file")
    file_name = "file_name_3"
    file = Path(file_name)
    artifacts = [Artifact(lims, id="2-2336715")]

    # WHEN running get_file_data_and_write
    # THEN MissingUDFsError is being raised and no file is created
    with pytest.raises(MissingUDFsError):
        get_file_data_and_write(artifacts, file)
    assert file.exists() is False



from genologics.entities import Process
from cg_lims.EPPs.udf.copy.reads_to_sequence import find_reads_to_sequence
from cg_lims.get.artifacts import get_artifacts
from cg_lims.get.samples import get_process_samples
from tests.conftest import server



def test_find_reads_to_sequence(lims):
    # GIVEN: A process with 5 input artifacts, where one of the input artifacts has the
    # udf 'Missing reads Pool (M)' set to 190.
    # And all the samples have udf 'Reads to sequence (M)' set to 6

    server("find_reads_to_sequence")
    process = Process(lims, id="24-196211")
    output_artifacts = get_artifacts(process=process, input=False)
    for output_artifact in output_artifacts:
        output_artifact.udf.clear()
        output_artifact.put()

    # WHEN running find_reads_to_sequence
    passed_artifacts, failed_artifacts = find_reads_to_sequence(process=process, lims=lims)

    # THEN the output artifact udf 'Reads to sequence (M)' will be fetched from the
    # input artifact udf 'Missing reads Pool (M)' if it exist,
    # oterwise from sample udf 'Reads to sequence (M)'
    assert passed_artifacts == 5
    assert failed_artifacts == 0
    for artifact in output_artifacts:
        if artifact.id == '2-1454725':
            assert artifact.udf['Reads to sequence (M)'] == '190'
        else:
            assert artifact.udf['Reads to sequence (M)'] == '6'


def test_find_reads_to_sequence_missing_udfs(lims):
    # GIVEN: A process with 5 input artifacts, where one of the input artifacts has the
    # udf 'Missing reads Pool (M)' set to 190
    # And no samples have udf 'Reads to sequence (M)' set.

    server("find_reads_to_sequence")
    process = Process(lims, id="24-196211")
    samples=get_process_samples(process=process)
    for sample in samples:
        sample.udf.clear()
        sample.put()

    output_artifacts = get_artifacts(process=process, input=False)

    # WHEN running find_reads_to_sequence
    passed_artifacts, failed_artifacts = find_reads_to_sequence(process=process, lims=lims)

    # THEN the output artifact udf 'Reads to sequence (M)' will be fetched from the
    # input artifact udf 'Missing reads Pool (M)' if it exist,
    # the other artifacts will fail.
    assert passed_artifacts == 1
    assert failed_artifacts == 4



#!/usr/bin/env python

import logging
import sys
from typing import Optional

import click

from cg_lims import options
from cg_lims.exceptions import LimsError
from cg_lims.get.artifacts import filter_artifacts, get_artifacts
from cg_lims.put.queue import queue_artifacts

LOG = logging.getLogger(__name__)


@click.command()
@options.workflow_id(help="Destination workflow id.")
@options.stage_id(help="Destination stage id.")
@options.udf(help="UDF that will tell which artifacts to move.")
@options.input(
    help="Use this flag if you want to queue the input artifacts of the current process. Default is to queue the "
    "output artifacts (analytes) of the process. "
)
@click.pass_context
def move_samples(ctx, workflow_id: str, stage_id: str, udf: Optional[str], input: bool):
    """Script to move aritfats to another stage.

    Queueing artifacts from the current process with <udf>==True, to stage with <stage-id>
    in workflow with <workflow-id>. Raising error if queuing fails.

    If udf is None all artifacts from the current step will be queued."""

    process = ctx.obj["process"]
    lims = ctx.obj["lims"]

    artifacts = get_artifacts(process=process, input=input)
    filtered_artifacts = filter_artifacts(artifacts, udf, True)

    if not filtered_artifacts:
        LOG.info("No artifacts to queue.")
        return

    try:
        queue_artifacts(lims, filtered_artifacts, workflow_id, stage_id)
        click.echo("Artifacts have been queued.")
    except LimsError as e:
        sys.exit(e.message)



from tests.conftest import server
import logging

from genologics.entities import Artifact

import pytest

from cg_lims.EPPs.udf.check.check_artifact_udfs import check_udfs
from cg_lims.exceptions import MissingUDFsError


def test_check_udfs_missing_udfs(lims):
    # GIVEN:
    server("flat_tests")

    artifacts = [Artifact(lims, id="1"), Artifact(lims, id="2")]
    artifact_udfs = ["Amount taken (ng)", "PCR Plate"]

    # WHEN running check_udfs
    # THEN Missing udfs error is raised
    with pytest.raises(MissingUDFsError):
        check_udfs(artifacts=artifacts, artifact_udfs=artifact_udfs)


def test_check_udfs(caplog, lims):
    # GIVEN:
    server("flat_tests")

    artifact_1 = Artifact(lims, id="1")
    artifact_2 = Artifact(lims, id="2")
    artifact_1.udf["Amount taken (ng)"] = 15
    artifact_1.udf["PCR Plate"] = "some plate"
    artifact_2.udf["Amount taken (ng)"] = 15
    artifact_2.udf["PCR Plate"] = "some plate"
    artifact_2.put()
    artifact_1.put()
    artifacts = [artifact_1, artifact_2]
    artifact_udfs = ["Amount taken (ng)", "PCR Plate"]

    # WHEN running check_udfs
    with caplog.at_level(logging.INFO):
        check_udfs(artifacts=artifacts, artifact_udfs=artifact_udfs)
        # THEN no error is being raised and logging id informing that artifacts were set
        assert "Artifact udfs were all set" in caplog.text



#!/usr/bin/env python
import click

from cg_lims.EPPs.move.move_samples import move_samples
from cg_lims.EPPs.move.place_samples_in_seq_agg import place_samples_in_seq_agg

# commands
from cg_lims.EPPs.move.rerun_samples import rerun_samples


@click.group(invoke_without_command=True)
@click.pass_context
def move(ctx):
    """Main entry point of move commands"""
    pass


move.add_command(rerun_samples)
move.add_command(move_samples)
move.add_command(place_samples_in_seq_agg)



"""
CLI module to calculate the resuspension buffer volume (RB Volume) based on the amount of DNA
needed, as specified in the step `CG002 - Normalization of microbial samples for sequencing`

For an explanation of the volume calculations, see AM-document 1764 Method - Manual
cost-efficient library preparation for microbial WGS """
import logging
import sys
from typing import List

import click
from genologics.entities import Artifact

from cg_lims.exceptions import HighConcentrationError, LimsError, MissingUDFsError
from cg_lims.get.artifacts import get_artifacts

FINAL_CONCENTRATION = 2
LOG = logging.getLogger(__name__)
MAXIMUM_CONCENTRATION = 60
QC_FAILED = "FAILED"
QC_PASSED = "PASSED"


def calculate_volumes(artifacts: List[Artifact]) -> None:
    """Determines the total volume, water volume, and sample volume"""
    missing_udfs = 0
    high_concentration = False
    qc_flag = QC_PASSED
    for artifact in artifacts:
        concentration = artifact.udf.get("Concentration")
        if concentration is None:
            LOG.error(
                f"Sample {artifact.samples[0].name} is missing udf 'Concentration'."
            )
            missing_udfs += 1
            continue
        if artifact.samples[0].name.startswith("NTC-CG"):
            sample_volume = 0
            buffer_volume = 15
        elif concentration < FINAL_CONCENTRATION:
            sample_volume = 15
            buffer_volume = 0
        elif 2 <= concentration <= 7.5:
            total_volume = 15
            sample_volume = float(FINAL_CONCENTRATION * total_volume) / concentration
            buffer_volume = total_volume - sample_volume
            if buffer_volume < 2:
                sample_volume = 15
                buffer_volume = 0
        elif 7.5 < concentration <= MAXIMUM_CONCENTRATION:
            sample_volume = 4
            total_volume = float(concentration * sample_volume) / FINAL_CONCENTRATION
            buffer_volume = total_volume - sample_volume
        else:
            high_concentration = True
            artifact.qc_flag = QC_FAILED
            artifact.put()
            continue
        artifact.qc_flag = qc_flag
        artifact.udf["Sample Volume (ul)"] = sample_volume
        artifact.udf["Volume Buffer (ul)"] = buffer_volume
        artifact.udf["Total Volume (uL)"] = buffer_volume + sample_volume
        artifact.put()

    if missing_udfs:
        raise MissingUDFsError(
            f"Could not apply calculations for {missing_udfs} out of {len(artifacts)} sample(s): "
            f"'Concentration' is missing!"
        )

    if high_concentration:
        mu = "\u03BC"
        raise HighConcentrationError(
            f"Could not apply calculations for one or more sample(s): concentration too high (> "
            f"60 ng/{mu}l)! "
        )


@click.command()
@click.pass_context
def calculate_microbial_aliquot_volumes(context: click.Context):
    """Calculate microbial aliquot volumes based on the Concentration udf"""
    LOG.info(f"Running {context.command_path} with params: {context.params}")
    process = context.obj["process"]
    try:
        artifacts: List[Artifact] = get_artifacts(process=process, input=False)
        calculate_volumes(artifacts=artifacts)
        message = "Microbial aliquot volumes have been calculated."
        LOG.info(message)
        click.echo(message)
    except LimsError as e:
        LOG.error(e.message)
        sys.exit(e.message)



"""CLI module to set Reads Missing (M) udf on new samples"""
import logging
import sys
from typing import List

import click
from genologics.entities import Sample

from cg_lims.exceptions import LimsError, MissingUDFsError
from cg_lims.get.samples import get_process_samples
from cg_lims.get.udfs import get_udf
from cg_lims.status_db_api import StatusDBAPI

LOG = logging.getLogger(__name__)


def get_target_amount(app_tag: str, status_db: StatusDBAPI) -> int:
    """Gets the target amount of reads from clinical-api"""
    return status_db.apptag(tag_name=app_tag, key="target_reads")


def set_reads_missing_on_sample(sample: Sample, status_db: StatusDBAPI) -> None:
    """Sets the udf "Reads missing (M)" on a sample to the target amount of reads bases on the app tag"""
    app_tag = get_udf(sample, "Sequencing Analysis")
    target_amount = get_target_amount(app_tag, status_db)
    sample.udf["Reads missing (M)"] = target_amount / 1000000
    sample.put()


def set_reads_missing(samples: List[Sample], status_db: StatusDBAPI) -> None:
    """Attempts to set the udf "Reads missing (M)" on all samples"""
    failed_samples_count = 0
    succeeded_samples_count = 0
    for sample in samples:
        try:
            set_reads_missing_on_sample(sample, status_db)
            succeeded_samples_count += 1
        except MissingUDFsError:
            failed_samples_count += 1

    if failed_samples_count:
        raise LimsError(
            f"Reads Missing (M) set on {succeeded_samples_count} sample(s), {failed_samples_count} sample(s) failed"
        )


@click.command()
@click.pass_context
def set_reads_missing_on_new_samples(context: click.Context):
    """Set Reads Missing (M) udf on samples before they go into any workflow"""

    LOG.info(f"Running {context.command_path} with params: {context.params}")

    process = context.obj["process"]
    status_db = context.obj["status_db"]

    samples = get_process_samples(process=process)

    try:
        set_reads_missing(samples, status_db)
        message = f"Reads Missing (M) udf set on all sample(s)."
        LOG.info(message)
        click.echo(message)
    except LimsError as e:
        LOG.error(e.message)
        sys.exit(e.message)



import csv
import sys
from pathlib import Path
from typing import List

import click
import logging
from genologics.lims import Lims

from cg_lims import options
from cg_lims.EPPs.files.csv_for_kapa_truble_shooting.models import DebugKapaCSV, HEADERS
from cg_lims.models.api.master_steps import (
    HybridizeLibraryTWIST,
    AliquotsamplesforenzymaticfragmentationTWIST,
    KAPALibraryPreparation,
    PoolsamplesforhybridizationTWIST,
    CaptureandWashTWIST,
    BeadPurificationTWIST,
    BufferExchange,
)
from cg_lims.exceptions import LimsError, FileError
from cg_lims.get.artifacts import get_artifact_by_name
from cg_lims.get.files import get_file_path

LOG = logging.getLogger(__name__)


def file_has_comma(file: Path) -> bool:
    """Checking for , in file."""
    with open(file, "r") as file_content:
        content_as_string = str(file_content.read())
        return "," in content_as_string


def build_sample_row(lims: Lims, sample_id: str) -> list:
    sample_row = DebugKapaCSV(SampleID=sample_id)
    sample_row.set_hybridize(hybridize=HybridizeLibraryTWIST(lims=lims, sample_id=sample_id))
    sample_row.set_aliquot(
        aliquot=AliquotsamplesforenzymaticfragmentationTWIST(lims=lims, sample_id=sample_id)
    )
    sample_row.set_kapa(kapa=KAPALibraryPreparation(lims=lims, sample_id=sample_id))
    sample_row.set_pool(pool=PoolsamplesforhybridizationTWIST(lims=lims, sample_id=sample_id))
    sample_row.set_capture(capture=CaptureandWashTWIST(lims=lims, sample_id=sample_id))
    sample_row.set_bead(bead=BeadPurificationTWIST(lims=lims, sample_id=sample_id))
    sample_row.set_buffer(buffer=BufferExchange(lims=lims, sample_id=sample_id))

    sample_row_dict = sample_row.dict(by_alias=True)
    return [sample_row_dict.get(header) for header in HEADERS]


@click.command()
@options.samples_file(help="Txt file with sample ids. (One per row!)")
@options.file_placeholder(help="File placeholder name.")
@options.local_file()
@click.pass_context
def trouble_shoot_kapa(ctx, samples_file: str, file: str, local_file: str):
    LOG.info(f"Running {ctx.command_path} with params: {ctx.params}")

    lims = ctx.obj["lims"]
    process = ctx.obj["process"]
    if local_file:
        file_path_str = local_file
    else:
        file_art = get_artifact_by_name(process=process, name=samples_file)
        file_path_str = get_file_path(file_art)

    file_path = Path(file_path_str)

    try:
        if not file_path.exists():
            LOG.error("File not existing")
            raise FileError(message="File not existing")

        if file_has_comma(file=file_path):
            LOG.error("File wrong format.")
            raise FileError(message="File wrong format.")

        with open(file_path_str, "r") as samples:
            sample_list = [sample_id.strip("\n") for sample_id in samples.readlines()]

        with open(f"{file}_kapa_debug.csv", "w", newline="\n") as new_csv:
            wr = csv.writer(new_csv, delimiter=",")
            wr.writerow(HEADERS)
            sample_rows: List[List[str]] = [
                build_sample_row(lims, sample_id) for sample_id in sample_list
            ]
            wr.writerows(sample_rows)
        click.echo("Twist csv file has been generated.")
    except LimsError as e:
        LOG.info(e.message)
        sys.exit(e.message)



from typing import Optional

from genologics.entities import Artifact, Process
from pydantic import Field, validator

from cg_lims.models.api.master_steps.base_step import (
    get_artifact_udf,
    BaseStep,
    get_process_udf,
)
from cg_lims.get.artifacts import get_latest_analyte


class KAPALibraryPreparation(BaseStep):
    "KAPA Library Preparation"
    artifact: Optional[Artifact]
    process: Optional[Process]

    label: Optional[str] = Field(None, alias="label (= index)")
    method_document: Optional[str] = Field(None, alias="Method document")
    document_version: Optional[int] = Field(None, alias="Document version")
    prep_kit: Optional[str] = Field(None, alias="KAPA HyperPlus/Prep Kit")
    well: Optional[str] = Field(None, alias="Well")
    container_name: Optional[str] = Field(None, alias="Container Name")
    size_bp: Optional[int]
    concentration: Optional[str]

    @validator("artifact", always=True)
    def get_artifact(cls, v, values):
        try:
            return get_latest_analyte(
                lims=values.get("lims"),
                sample_id=values.get("sample_id"),
                process_types=["KAPA Library Preparation TWIST v1"],
            )
        except:
            return None

    @validator("process", always=True)
    def get_process(cls, v, values):
        if values.get("artifact"):
            return values.get("artifact").parent_process
        return None

    @validator("method_document", always=True)
    def get_method_document(cls, v, values):
        return get_process_udf(values.get("process"), "Method document")

    @validator("document_version", always=True)
    def get_document_version(cls, v, values):
        return get_process_udf(values.get("process"), "Document version")

    @validator("prep_kit", always=True)
    def get_prep_kit(cls, v, values):
        return get_process_udf(values.get("process"), "KAPA HyperPlus/Prep Kit")

    @validator("container_name", always=True)
    def get_container_name(cls, v, values):
        artifact = values.get("artifact")
        if artifact:
            return artifact.container.name
        return None

    @validator("well", always=True)
    def get_well(cls, v, values):
        artifact = values.get("artifact")
        if artifact:
            return artifact.location[1]
        return None

    @validator("label", always=True)
    def get_kapa_label(cls, v, values):
        artifact = values.get("artifact")
        if artifact:
            return artifact.reagent_labels
        return None

    @validator("size_bp", always=True)
    def get_kapa_size_bp(cls, v, values):
        return get_artifact_udf(values.get("artifact"), "Size (bp)")

    @validator("concentration", always=True)
    def get_kapa_concentration(cls, v, values):
        return get_artifact_udf(values.get("artifact"), "Concentration")



import pytest
from genologics.entities import Artifact, Process

from cg_lims.EPPs.udf.copy.artifact_to_artifact import copy_udfs_to_all_artifacts
from cg_lims.get.artifacts import get_artifacts, get_latest_analyte
from cg_lims.exceptions import MissingUDFsError
from tests.conftest import server


def test_copy_udf(lims):
    # GIVEN an Aggregate QC process with samples which have source artifacts that contain the
    # Amount (ng) UDF
    server("wgs_prep")
    process = Process(lims, id="24-240276")
    artifacts = get_artifacts(process=process, input=True)

    # WHEN running get_copy_tasks to copy the Amount (ng) UDF
    copy_udfs_to_all_artifacts(
        artifacts=artifacts,
        udfs=[("Amount (ng)", "Amount (ng)")],
        process_types=["Buffer Exchange v2"],
        lims=lims,
        sample_artifact=True,
        qc_flag=True,
        ignore_fail=False,
        keep_failed_flags=False,
    )

    # THEN the correct source UDFs are copied over
    for artifact in artifacts:
        sample_id = artifact.samples[0].id
        source_artifact = Artifact(lims, id=f"{sample_id}PA1")
        assert artifact.udf["Amount (ng)"] == source_artifact.udf["Amount (ng)"]


def test_copy_udf_process_type(lims):
    # GIVEN an Aggregate QC process with samples which have passed through an Aliquot Samples for Covaris step
    server("wgs_prep")
    process = Process(lims, id="24-240276")
    artifacts = get_artifacts(process=process, input=True)

    # WHEN running get_copy_tasks to copy the Concentration UDF
    copy_udfs_to_all_artifacts(
        artifacts=artifacts,
        udfs=[("Concentration", "Concentration")],
        process_types=["Aliquot Samples for Covaris"],
        lims=lims,
        sample_artifact=True,
        qc_flag=True,
        ignore_fail=False,
        keep_failed_flags=False,
    )

    # THEN the correct source UDFs are copied over
    for artifact in artifacts:
        sample_id = artifact.samples[0].id
        source_artifact = get_latest_analyte(lims,
                                             sample_id=sample_id,
                                             process_types=["Aliquot Samples for Covaris"],
                                             sample_artifact=False)
        assert artifact.udf["Concentration"] == source_artifact.udf["Concentration"]


def test_copy_udf_no_source(lims):
    # GIVEN an Aggregate QC process with samples which have source artifacts that doesn't contain the
    # DIN Value UDF
    server("wgs_prep")
    process = Process(lims, id="24-240276")
    artifacts = get_artifacts(process=process, input=True)

    # WHEN running get_copy_tasks
    # THEN MissingUDFError is raised
    with pytest.raises(MissingUDFsError) as error_message:
        copy_udfs_to_all_artifacts(
            artifacts=artifacts,
            udfs=[("DIN Value", "DIN Value")],
            process_types=["Buffer Exchange v2"],
            lims=lims,
            sample_artifact=True,
            qc_flag=True,
            ignore_fail=False,
            keep_failed_flags=False,
        )


def test_copy_udf_no_source_ignore(lims):
    # GIVEN an Aggregate QC process with samples which have source artifacts that doesn't contain the
    # DIN Value UDF
    server("wgs_prep")
    process = Process(lims, id="24-240276")
    artifacts = get_artifacts(process=process, input=True)

    # WHEN running get_copy_tasks with ignore_fail set to true
    # THEN no MissingUDFError is raised
    copy_udfs_to_all_artifacts(
        artifacts=artifacts,
        udfs=[("DIN Value", "DIN Value")],
        process_types=["Buffer Exchange v2"],
        lims=lims,
        sample_artifact=True,
        qc_flag=True,
        ignore_fail=True,
        keep_failed_flags=False,
    )


def test_copy_qc_flag(lims):
    # GIVEN an Aggregate QC process with samples which have source artifacts with either passing or failed qc flags
    server("wgs_prep")
    process = Process(lims, id="24-240276")
    artifacts = get_artifacts(process=process, input=True)
    source_artifact = Artifact(lims, id="ACC8454A1PA1")
    source_artifact.qc_flag = 'FAILED'
    source_artifact.put()

    # WHEN running get_copy_tasks with qc_flag set to true
    copy_udfs_to_all_artifacts(
        artifacts=artifacts,
        udfs=[],
        process_types=[],
        lims=lims,
        sample_artifact=True,
        qc_flag=True,
        ignore_fail=False,
        keep_failed_flags=False,
    )

    # THEN the correct QC flags are copied over
    for artifact in artifacts:
        sample_id = artifact.samples[0].id
        source_artifact = Artifact(lims, id=f"{sample_id}PA1")
        assert artifact.qc_flag == source_artifact.qc_flag


def test_copy_qc_flag_fail(lims):
    # GIVEN an Aggregate QC process with samples with mixed failed and passed qc flags
    server("wgs_prep")
    process = Process(lims, id="24-240276")
    artifacts = get_artifacts(process=process, input=True)
    artifacts[0].qc_flag = 'FAILED'
    artifacts[0].put()

    # WHEN running get_copy_tasks
    copy_udfs_to_all_artifacts(
        artifacts=artifacts,
        udfs=[],
        process_types=[],
        lims=lims,
        sample_artifact=True,
        qc_flag=True,
        ignore_fail=False,
        keep_failed_flags=True,
    )

    # THEN failed qc flags are not overwritten
    assert artifacts[0].qc_flag == 'FAILED'
    for artifact in artifacts[1:5]:
        sample_id = artifact.samples[0].id
        source_artifact = Artifact(lims, id=f"{sample_id}PA1")
        assert artifact.qc_flag == source_artifact.qc_flag



from typing import Optional

from genologics.lims import Lims
from pydantic import Field, BaseModel

from cg_lims.objects import BaseAnalyte
from cg_lims.models.arnold.base_step import BaseStep


class ProcessUDFs(BaseModel):
    hybridization_time: Optional[float] = Field(None, alias="Total hybridization time (h)")
    lot_nr_h20_capture_and_wash: Optional[str] = Field(None, alias="Nuclease free water")
    capture_and_wash_method: Optional[str] = Field(None, alias="Method document")
    binding_and_purification_beads: Optional[str] = Field(
        None, alias="Twist Binding and Purification beads"
    )
    wash_buffers: Optional[str] = Field(None, alias="Twist Wash Buffers")


class ArnoldStep(BaseStep):
    process_udfs: ProcessUDFs

    class Config:
        allow_population_by_field_name = True


def get_capture_and_wash(lims: Lims, sample_id: str, prep_id: str) -> ArnoldStep:
    analyte = BaseAnalyte(
        lims=lims,
        sample_id=sample_id,
        process_type="Target enrichment TWIST v1",
    )

    return ArnoldStep(
        **analyte.base_fields(),
        process_udfs=ProcessUDFs(**analyte.process_udfs()),
        sample_id=sample_id,
        prep_id=prep_id,
        step_type="capture_and_wash",
        workflow="TWIST",
    )



import logging
import pathlib
from typing import List

from genologics.config import BASEURI
from genologics.entities import Artifact, Process, Sample
from genologics.lims import Lims

from cg_lims.get.ids import unique_list_of_ids

LOG = logging.getLogger(__name__)

from cg_lims.exceptions import MissingArtifactError, QueueArtifactsError


def queue_artifacts(lims: Lims, artifacts: List[Artifact], workflow_id: str, stage_id: str) -> None:
    """Queue artifacts to stage in workflow."""

    if not artifacts:
        LOG.warning("Failed trying to queue empty list of artifacts.")
        raise MissingArtifactError("No artifacts to queue.")
    stage_uri = f"{BASEURI}/api/v2/configuration/workflows/{workflow_id}/stages/{stage_id}"

    artifact_ids = unique_list_of_ids(artifacts)
    lims.route_artifacts(artifacts, stage_uri=stage_uri)
    try:
        lims.route_artifacts(artifacts, stage_uri=stage_uri)
        LOG.info(f"Queueing artifacts to {stage_uri}.")
        LOG.info(f"The following artifacts have been queued: {' ,'.join(artifact_ids)}")
    except:
        raise QueueArtifactsError("Failed to queue artifacts.")



from statistics import mean
from typing import Optional, List

from genologics.lims import Lims, Process, Artifact
from pydantic.main import BaseModel
from pydantic import Field

from cg_lims.get.artifacts import get_artifacts
from cg_lims.models.arnold.base_step import BaseStep


class ProcessUDFs(BaseModel):
    """Bcl Conversion & Demultiplexing (Nova Seq)"""

    methods: Optional[str] = Field(None, alias="Methods")
    atlas_version: Optional[str] = Field(None, alias="Atlas Version")
    pct_Q30_bases_threshold: Optional[str] = Field(None, alias="Threshold for % bases >= Q30")


class ArtifactUDFs(BaseModel):
    pct_bases_over_q30: Optional[float]
    number_of_reads: Optional[float]


class ArnoldStep(BaseStep):
    process_udfs: ProcessUDFs
    artifact_udfs: ArtifactUDFs

    class Config:
        allow_population_by_field_name = True


def get_sum_reads(sum_reads_udf: str, artifacts: List[Artifact]) -> Optional[int]:
    sum_reads = 0
    for artifact in artifacts:
        try:
            sum_reads += int(artifact.udf.get(sum_reads_udf))
        except ValueError:
            return None

    return sum_reads


def get_average_q30(q30_udf: str, artifacts: List[Artifact]) -> Optional[float]:
    q_30 = []
    for artifact in artifacts:
        try:
            q_30.append(float(artifact.udf.get(q30_udf)))
        except ValueError:
            return None
    return mean(q_30)


def get_bcl_conversion_and_demultiplexing(
    lims: Lims, sample_id: str, prep_id: str, process_id: str
) -> ArnoldStep:
    process = Process(lims, id=process_id)
    process.get()
    artifacts = get_artifacts(process=process, reagent_label=True)
    return ArnoldStep(
        process_udfs=ProcessUDFs(**dict(process.udf.items())),
        artifact_udfs=ArtifactUDFs(
            pct_bases_over_q30=get_average_q30(artifacts=artifacts, q30_udf="% Bases >=Q30"),
            number_of_reads=get_sum_reads(artifacts=artifacts, sum_reads_udf="# Reads"),
        ),
        lims_step_name=process.type.name,
        date_run=process.date_run,
        sample_id=sample_id,
        prep_id=prep_id,
        step_type="bcl_conversion_and_demultiplexing",
        workflow="NovaSeq",
    )



import logging
import sys
from typing import List, Tuple
from datetime import date
import click
from genologics.entities import Artifact, Process
from cg_lims import options
from cg_lims.exceptions import LimsError
from cg_lims.get.artifacts import get_artifacts
from cg_lims.EPPs.files.pooling_map.hmtl_templates import (
    PLACEMENT_MAP_HEADER,
    POOL_HEADER,
    SAMPLE_COLUMN_HEADERS,
    SAMPLE_COLUMN_VALUES,
)
from cg_lims.EPPs.files.pooling_map.models import (
    PlacementMapHeader,
    PoolSection,
    SampleTableSection,
)
from matplotlib import colors

COLORS = list(colors.TABLEAU_COLORS.values())
LOG = logging.getLogger(__name__)


def add_pool_info(pool_udfs: List[str], pool: Artifact) -> str:
    """Adding info about the pool"""

    html = []
    for udf in pool_udfs:
        value = pool.udf.get(udf)
        if value is not None:
            html.append(
                f'<tr><td class="group-field-label">{udf}: </td><td class="group-field-value">{value}</td></tr>'
            )
    return "".join(html)


def add_sample_info_headers(udfs_headers: List[str]) -> str:
    """Adding headers for more info about samples in the pool"""

    html = [f'<th style="width: 7%;" class="">{header}</th>' for header in udfs_headers]

    return "".join(html)


def add_sample_info(artifact: Artifact, udfs: List[str]) -> str:
    """Adding info about samples in the pool"""

    html = []
    for udf in udfs:
        value = artifact.udf.get(udf, "")
        html.append(f'<td class="" style="width: 7%;">{value}</td>')
    return "".join(html)


def make_html(
    pools: List[Artifact],
    process: Process,
    pool_udfs: List[str],
    sample_udfs: List[str],
) -> str:
    """Building the html for the pooling map"""

    html = []
    header_info = PlacementMapHeader(process_type=process.type.name, date=date.today().isoformat())
    html.append(PLACEMENT_MAP_HEADER.format(**header_info.dict()))

    for pool in pools:
        artifacts: List[Tuple[str, Artifact]] = [
            (artifact.location[1], artifact) for artifact in pool.input_artifact_list()
        ]
        # sorting on columns
        artifacts.sort(key=lambda tup: tup[0][1])

        pool_info = PoolSection(
            nr_samples=len(artifacts),
            pool_name=pool.name,
            pool_id=pool.id,
            pools_information=add_pool_info(pool_udfs, pool),
        )
        html.append(POOL_HEADER.format(**pool_info.dict()))
        extra_sample_columns: str = add_sample_info_headers(sample_udfs)
        html.append(SAMPLE_COLUMN_HEADERS.format(extra_sample_columns=extra_sample_columns))
        html.append("""</thead><tbody>""")
        source_containers = []
        for location, artifact in artifacts:
            sample = artifact.samples[0]
            sample_warning_color = "#F08080" if artifact.udf.get("Warning") else "#FFFFFF"
            if artifact.container.name not in source_containers:
                source_container_color = COLORS[0]
                COLORS.pop(0)
            source_containers.append(artifact.container.name)
            sample_table_values = SampleTableSection(
                sample_id=sample.id,
                sample_warning_color=sample_warning_color,
                source_well=location,
                source_container=artifact.container.name,
                source_container_color=source_container_color,
                pool_name=pool.name,
                extra_sample_values=add_sample_info(artifact, sample_udfs),
            )
            html.append(SAMPLE_COLUMN_VALUES.format(**sample_table_values.dict()))
        html.append("""</tbody></table><br><br></html>""")

    html = "".join(html)
    return html


@click.command()
@options.file_placeholder(help="File placeholder name.")
@options.pool_udfs(help="Pool UDFs to show in the placement map.")
@options.sample_udfs(help="Sample UDFs to show in the placement map.")
@click.pass_context
def pool_map(ctx, file: str, sample_udfs: List[str], pool_udfs: List[str]):
    """Create a pool placement map."""

    LOG.info(f"Running {ctx.command_path} with params: {ctx.params}")
    process = ctx.obj["process"]

    try:
        pools = get_artifacts(process=process, input=False)
        html = make_html(
            pools,
            process,
            pool_udfs,
            sample_udfs,
        )
        with open(f"{file}.html", "w") as file:
            file.write(html)
        click.echo("The file was successfully generated.")
    except LimsError as e:
        sys.exit(e.message)



#!/usr/bin/env python
import logging
import sys

import click
from cg_lims.get.processes import get_latest_process
from genologics.entities import Process, Artifact
from genologics.lims import Lims

from cg_lims.exceptions import LimsError, MissingUDFsError
from cg_lims.get.artifacts import get_artifacts
from typing import List, Dict, Literal, Any


def get_copy_tasks(
    process: Process,
) -> Dict[str, Dict[Literal["Source Step", "Source Field"], Any]]:
    copy_tasks = {}

    for udf, value in process.udf.items():
        if "Copy task" not in udf:
            continue
        task_number, copy_type = udf.split("-")
        task_number = task_number.strip()
        copy_type = copy_type.strip()
        if copy_type not in ["Source Step", "Source Field"]:
            raise MissingUDFsError(
                f'Copy type was: {copy_type}! Must be "Source Step" or "Source Field"'
            )
        if task_number in copy_tasks:
            copy_tasks[task_number][copy_type] = value
        else:
            copy_tasks[task_number] = {copy_type: value}
    if copy_tasks == {}:
        raise MissingUDFsError(f"Copy task UDF missing for process {process.id}")
    return copy_tasks


def copy_udfs(input_artifacts: List[Artifact], copy_tasks: dict, lims: Lims):
    """Loop through all artifacts and copy udfs from the corect steps."""
    failed_udfs = []
    process_types = [task["Source Step"] for task_number, task in copy_tasks.items()]
    process_types = list(set(process_types))
    for artifact in input_artifacts:
        artifact.get()
        source_steps = get_correct_steps_for_the_artifact(
            process_types=process_types, artifact=artifact, lims=lims
        )
        for process_type, source_step in source_steps.items():
            for output in source_step.outputs_per_input(artifact.id):
                qc_flag_update = Artifact(lims, id=output.id).qc_flag
                if not qc_flag_update or qc_flag_update == "UNKNOWN":
                    continue
                if artifact.qc_flag != "FAILED":
                    artifact.qc_flag = qc_flag_update
                for task_number, task in copy_tasks.items():
                    if task["Source Step"] != process_type:
                        continue
                    udf = task["Source Field"]
                    udf_value = output.udf.get(udf)
                    if udf_value is None:
                        failed_udfs.append(artifact.name)
                        continue
                    try:
                        artifact.udf[udf] = float(udf_value)
                    except:
                        artifact.udf[udf] = str(udf_value)

        artifact.put()


def get_correct_steps_for_the_artifact(
    process_types: List[str], artifact: Artifact, lims: Lims
) -> Dict[str, Process]:
    """Get the latest processes of the specified process types."""

    source_steps = {}
    processes = lims.get_processes(inputartifactlimsid=artifact.id, type=process_types)
    for process in processes:
        process_type = process.type.name
        if process_type in source_steps:
            source_steps[process_type] = get_latest_process([source_steps[process_type], process])
        else:
            source_steps[process_type] = process
    return source_steps


@click.command()
@click.pass_context
def aggregate_qc_and_copy_fields(ctx):
    """Script to copy artifact udf to sample udf"""

    process = ctx.obj["process"]
    lims = ctx.obj["lims"]
    try:
        copy_tasks = get_copy_tasks(process=process)
        artifacts = get_artifacts(process=process, input=True)
        copy_udfs(input_artifacts=artifacts, copy_tasks=copy_tasks, lims=lims)
        click.echo("Udfs have been set on all samples.")
    except LimsError as e:
        sys.exit(e.message)



#!/usr/bin/env python

from setuptools import find_packages, setup

from cg_lims import __version__ as version

try:
    with open("requirements.txt", "r") as f:
        install_requires = [x.strip() for x in f.readlines()]
except IOError:
    install_requires = []

with open("README.md", "r") as fh:
    long_description = fh.read()


setup(
    name="cg_lims",
    version=version,
    description="Lims code for Clinical Genomics",
    long_description=long_description,
    long_description_content_type="text/markdown",
    author="Maya Brandi",
    url="https://github.com/Clinical-Genomics/cg_lims",
    author_email="maya.brandi@scilifelab.se",
    packages=find_packages(),
    include_package_data=True,
    entry_points={
        "console_scripts": ["lims=cg_lims.commands:cli"],
    },
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    python_requires=">=3.6", # genologics # requires python < 3.9  - AttributeError: 'xml.etree.ElementTree.Element' object has no attribute 'getchildren'
)



import logging
from typing import List

import click
from genologics.lims import Lims
import requests
from requests import Response
import json

from cg_lims import options
from cg_lims.EPPs.arnold.flow_cell import build_flow_cell_document
from cg_lims.get.artifacts import get_output_artifacts
from cg_lims.models.arnold.flow_cell import FlowCell

LOG = logging.getLogger(__name__)


@click.command()
@options.process_types()
@click.pass_context
def update_arnold_flow_cells(ctx, process_types: List[str]):
    """For ALL runs defined by process_types, updating ALL Flow Cell documents.
    This script should in other words be run with care."""

    LOG.info(f"Running {ctx.command_path} with params: {ctx.params}")

    lims: Lims = ctx.obj["lims"]
    arnold_host: str = ctx.obj["arnold_host"]
    processes = lims.get_processes(type=process_types)
    LOG.info(f"loading {len(processes)} processes")
    for process in processes:
        try:
            lanes = get_output_artifacts(
                process=process,
                output_generation_type=["PerInput"],
                lims=lims,
                output_type="ResultFile",
            )
            flow_cell_document: FlowCell = build_flow_cell_document(process=process, lanes=lanes)
            response: Response = requests.post(
                url=f"{arnold_host}/flow_cell",
                headers={"Content-Type": "application/json"},
                data=flow_cell_document.json(exclude_none=True),
            )

            if not response.ok:
                LOG.error(response.text)
            else:
                LOG.info(f"process loaded: {process.id}")
        except:
            LOG.error(f"faild with process {process.id}")

    click.echo("Done. See log file for details.")



import logging
import click
import pathlib
import yaml
from genologics.lims import Lims

from cg_lims import options

LOG = logging.getLogger(__name__)


@click.command()
@options.config()
@options.log()
def process_to_artifact(
    config: str,
    log: str,
):
    """Script to copy"""

    log_path = pathlib.Path(log)
    log_level = getattr(logging, "INFO")
    logging.basicConfig(filename=str(log_path.absolute()), filemode="w", level=log_level)

    with open(config) as file:
        config_data = yaml.load(file, Loader=yaml.FullLoader)
    lims = Lims(config_data["BASEURI"], config_data["USERNAME"], config_data["PASSWORD"])
    master_step = "Amplify Captured Libraries TWIST v2"
    artifacts = lims.get_artifacts(type="Analyte", process_type=master_step)
    for artifact in artifacts:
        process = artifact.parent_process
        pcr_cycles = process.udf.get("Nr of PCR cycles")
        if pcr_cycles is not None:
            artifact.udf["Nr of PCR cycles"] = str(pcr_cycles)
            artifact.put()
            print(artifact.id)
        else:
            LOG.error(f"process: {process.id} is missing process udf Nr of PCR cycles")


if __name__ == "__main__":
    process_to_artifact()



from typing import Optional

from genologics.lims import Lims
from pydantic.main import BaseModel
from pydantic import Field

from cg_lims.models.arnold.base_step import BaseStep
from cg_lims.objects import BaseAnalyte


class ProcessUDFs(BaseModel):
    """STANDARD Make Pool and Denature (Nova Seq)"""

    volume_of_pool: Optional[str] = Field(None, alias="Volume of Pool to Denature (ul")
    tris_lot_nr: Optional[str] = Field(None, alias="Tris-HCl lot nr")
    naoh_lot_nr: Optional[str] = Field(None, alias="NaOH lot nr")
    phix_lot_nr: Optional[str] = Field(None, alias="PhiX lot nr")
    tris_volume: Optional[str] = Field(None, alias="Tris-HCl Volume (ul)")
    naoh_volume: Optional[str] = Field(None, alias="NaOH Volume (ul)")
    phix_volume: Optional[str] = Field(None, alias="PhiX Volume (ul)")


class ArnoldStep(BaseStep):
    process_udfs: ProcessUDFs

    class Config:
        allow_population_by_field_name = True


def get_standard_make_pool_and_denature(
    lims: Lims, sample_id: str, prep_id: str
) -> Optional[ArnoldStep]:
    analyte = BaseAnalyte(
        lims=lims,
        sample_id=sample_id,
        process_type="STANDARD Make Pool and Denature (Nova Seq)",
        optional_step=True,
    )
    if not analyte.process:
        return None

    return ArnoldStep(
        **analyte.base_fields(),
        process_udfs=ProcessUDFs(**analyte.process_udfs()),
        sample_id=sample_id,
        prep_id=prep_id,
        step_type="standard_make_pool_and_denature",
        workflow="NovaSeq"
    )



from typing import List

from genologics.lims import Lims
from .buffer_exchange import get_buffer_exchange_twist
from .capture_and_wash_twist import get_capture_and_wash
from .pool_samples_twist import get_pool_samples_twist
from .hybridize_library_twist import get_hybridize_library_twist
from .bead_purification_twist import get_bead_purification_twist
from .kapa_library_preparation_twist import get_kapa_library_preparation_twist
from .aliquot_samples_for_enzymatic_fragmentation_twist import (
    get_aliquot_samples_for_enzymatic_fragmentation,
)
from .amplify_captured_libraries import get_amplify_captured_library_udfs
from .enzymatic_fragmentation_twist import get_enzymatic_fragmentation
from .reception_control import get_sample_artifact_fields
from cg_lims.models.arnold.base_step import BaseStep
from .normalization_of_samples_for_sequencing import get_normalization_of_samples


def build_twist_documents(sample_id: str, process_id: str, lims: Lims) -> List[BaseStep]:
    """Building a sars_cov_2 Prep."""
    prep_id = f"{sample_id}_{process_id}"
    step_documents = [
        get_aliquot_samples_for_enzymatic_fragmentation(
            sample_id=sample_id, lims=lims, prep_id=prep_id
        ),
        get_hybridize_library_twist(sample_id=sample_id, lims=lims, prep_id=prep_id),
        get_pool_samples_twist(sample_id=sample_id, lims=lims, prep_id=prep_id),
        get_capture_and_wash(sample_id=sample_id, lims=lims, prep_id=prep_id),
        get_kapa_library_preparation_twist(sample_id=sample_id, lims=lims, prep_id=prep_id),
        get_buffer_exchange_twist(sample_id=sample_id, lims=lims, prep_id=prep_id),
        get_bead_purification_twist(sample_id=sample_id, lims=lims, prep_id=prep_id),
        get_enzymatic_fragmentation(sample_id=sample_id, lims=lims, prep_id=prep_id),
        get_amplify_captured_library_udfs(sample_id=sample_id, lims=lims, prep_id=prep_id),
        get_sample_artifact_fields(sample_id=sample_id, lims=lims, prep_id=prep_id),
        get_normalization_of_samples(sample_id=sample_id, lims=lims, prep_id=prep_id),
    ]
    return [document for document in step_documents if document is not None]



import logging
from typing import List, Literal

import click
from genologics.lims import Lims
import requests
from requests import Response
import json

from cg_lims import options
from cg_lims.EPPs.arnold import prep, sequencing
from cg_lims.models.arnold.base_step import BaseStep

LOG = logging.getLogger(__name__)


@click.command()
@options.prep(help="Prep type.")
@options.process_types()
@click.pass_context
def update_arnold_preps(
    ctx, prep_type: Literal["wgs", "twist", "micro", "cov", "rna"], process_types: List[str]
):
    """For ALL preps defined by prep_type and process_types, updating ALL Step documents.
    This script should in other words be run with care."""

    LOG.info(f"Running {ctx.command_path} with params: {ctx.params}")

    lims: Lims = ctx.obj["lims"]
    arnold_host: str = ctx.obj["arnold_host"]
    processes = lims.get_processes(type=process_types)
    LOG.info(f"loading {len(processes)} processes")
    for process in processes:
        try:
            all_step_documents: List[BaseStep] = prep.build_step_documents(
                prep_type=prep_type, process=process, lims=lims
            )
            response: Response = requests.put(
                url=f"{arnold_host}/steps",
                headers={"Content-Type": "application/json"},
                data=json.dumps([doc.dict(exclude_none=True) for doc in all_step_documents]),
            )
            if not response.ok:
                LOG.error(response.text)
            else:
                LOG.info(f"process loaded: {process.id}")
        except:
            LOG.error(f"faild with process {process.id}")

    click.echo("Done. See log file for details.")


@click.command()
@options.sequencing_method(help="Sequencing Method.")
@options.process_types()
@click.pass_context
def update_arnold_runs(ctx, sequencing_method: Literal["novaseq"], process_types: List[str]):
    """For ALL sequencing runs defined by sequencing_method and process_types, updating ALL Step documents.
    This script should in other words be run with care."""

    LOG.info(f"Running {ctx.command_path} with params: {ctx.params}")

    lims: Lims = ctx.obj["lims"]
    arnold_host: str = ctx.obj["arnold_host"]
    processes = lims.get_processes(type=process_types)
    LOG.info(f"loading {len(processes)} processes")
    for process in processes:
        try:
            all_step_documents: List[BaseStep] = sequencing.build_step_documents(
                sequencing_method=sequencing_method, process=process, lims=lims
            )
            response: Response = requests.put(
                url=f"{arnold_host}/steps",
                headers={"Content-Type": "application/json"},
                data=json.dumps([doc.dict(exclude_none=True) for doc in all_step_documents]),
            )
            if not response.ok:
                LOG.error(response.text)
            else:
                LOG.info(f"process loaded: {process.id}")
        except:
            LOG.error(f"faild with process {process.id}")

    click.echo("Done. See log file for details.")



from typing import List

from genologics.entities import Artifact, Process, Sample

from cg_lims.exceptions import MissingSampleError
import logging

LOG = logging.getLogger(__name__)


def get_process_samples(process: Process) -> List[Sample]:
    """Get all samples in a process"""

    all_samples = []
    for art in process.all_inputs():
        all_samples += art.samples

    return list(set(all_samples))


def get_one_sample_from_artifact(artifact: Artifact) -> Sample:
    """Checking that an artifact has one and only one sample.
    Returning the sample if it exists.
    Raising MissingSampleError otherwise."""

    no_samples_message = f"Artifact {artifact.id} has no samples."
    more_than_one_message = f"Artifact {artifact.id} has more than one sample."
    try:
        samples = artifact.samples
    except:
        LOG.error(no_samples_message)
        raise MissingSampleError(message=no_samples_message)
    if not samples:
        LOG.error(no_samples_message)
        raise MissingSampleError(message=no_samples_message)
    elif len(samples) > 1:
        LOG.error(more_than_one_message)
        raise MissingSampleError(message=more_than_one_message)

    return samples[0]



from typing import Optional
from genologics.lims import Lims
from pydantic import Field, BaseModel
from cg_lims.objects import BaseAnalyte
from cg_lims.models.arnold.base_step import BaseStep


class ProcessUDFs(BaseModel):
    amplify_captured_library_method: Optional[str] = Field(None, alias="Method document")
    lot_nr_xgen_primer_amplify_captured_library: Optional[str] = Field(
        None, alias="xGen Library Amp primer"
    )
    lot_nr_amplification_kit_amplify_captured_library: Optional[str] = Field(
        None, alias="Kapa HiFi HotStart ReadyMix"
    )


class ArtifactUDFs(BaseModel):
    nr_pcr_cycles_amplify_captured_library: Optional[int] = Field(None, alias="Nr of PCR cycles")


class ArnoldStep(
    BaseStep,
):
    process_udfs: ProcessUDFs
    artifact_udfs: ArtifactUDFs

    class Config:
        allow_population_by_field_name = True


def get_amplify_captured_library_udfs(lims: Lims, sample_id: str, prep_id: str) -> ArnoldStep:
    analyte = BaseAnalyte(
        lims=lims,
        sample_id=sample_id,
        process_type="Target enrichment TWIST v1",
    )

    return ArnoldStep(
        **analyte.base_fields(),
        process_udfs=ProcessUDFs(**analyte.process_udfs()),
        artifact_udfs=ArtifactUDFs(**analyte.artifact_udfs()),
        sample_id=sample_id,
        prep_id=prep_id,
        step_type="amplify_captured_library",
        workflow="TWIST",
    )



#!/usr/bin/env python

import click

from .load_arnold_steps import update_arnold_preps, update_arnold_runs
from .load_arnold_flowcells import update_arnold_flow_cells


@click.group(invoke_without_command=True)
@click.pass_context
def one_time(ctx):
    """Main load commands."""
    pass


one_time.add_command(update_arnold_preps)
one_time.add_command(update_arnold_runs)
one_time.add_command(update_arnold_flow_cells)
